from tweepy.streaming import StreamListener
import json
import time


class Tweet(object):
    """
    Tweet object model.
    """
    def __init__(self, data):
        self._data = data
        self._user_keys = (
            'followers_count', 'friends_count', 'location'
        )
        try:
            ts = time.strftime(
                '%Y-%m-%d %H:%M:%S',
                time.strptime(data['created_at'], '%a %b %d %H:%M:%S +0000 %Y')
                )
            data['created_at'] = str(ts)
        except KeyError:
            print(data)
            raise

    def __len__(self):
        return len(self.keys())

    def __contains__(self, other):
        return other in self.keys()

    def __getitem__(self, key):
        if key in self._user_keys:
            return self._data['user'][key]
        else:
            return self._data[key]

    def keys(self):
        k = tuple(self._data.keys())
        if 'user' in self._data:
            k += self._user_keys
        return k


class TweetListener(StreamListener):
    def __init__(self, on_data, on_error):
        self._on_data = on_data
        self._on_error = on_error

    def on_data(self, data):
        dat = json.loads(data)
        if isinstance(dat, list):
            for d in dat:
                if 'created_at' in d:
                    twt = Tweet(d)
                    if hasattr(self._on_data, '__call__'):
                        self._on_data(twt)
                else:
                    continue
        else:
            if 'created_at' in dat:
                twt = Tweet(dat)
                if hasattr(self._on_data, '__call__'):
                    self._on_data(twt)
        return True

    def on_error(self, status):
        print(status)
        ret = True
        if hasattr(self._on_error, '__call__'):
            ret = self._on_error(status)
        if status == 420:
            #returning False in on_data disconnects the stream
            return False
        return ret



import time
from threading import Thread
from nltk.sentiment.vader import SentimentIntensityAnalyzer
from re import sub
from tweepy import OAuthHandler
from tweepy import Stream
from tweetfeels import TweetData
from tweetfeels import TweetListener


def clean(text):
    text = text.lower()
    text = sub("[0-9]+", "number", text)
    text = sub("#", "", text)
    text = sub("\n", "", text)
    text = text.replace('$', '@')
    text = sub("@[^\s]+", "", text)
    text = sub("(http|https)://[^\s]*", "", text)
    text = sub("[^\s]+@[^\s]+", "", text)
    text = sub('[^a-z A-Z]+', '', text)
    return text


class TweetFeels(object):
    """
    The controller.

    :param credentials: A list of your 4 credential components.
    :param tracking: A list of keywords to track.
    :param db: A sqlite database to store data. Will be created if it doesn't
               already exist. Will append if it exists.
    :ivar calc_every_n: Wont calculate new sentiment until there are n records
                        in the queue.
    :ivar lang: A list of languages to include in tweet gathering.
    """
    def __init__(self, credentials, tracking=[], db='feels.sqlite'):
        self._listener = TweetListener(self.on_data, self.on_error)
        self._feels = TweetData(db)
        _auth = OAuthHandler(credentials[0], credentials[1])
        _auth.set_access_token(credentials[2], credentials[3])
        self._stream = Stream(_auth, self._listener)
        self.tracking = tracking
        self.lang = ['en']
        self._sentiment = 0
        self._filter_level = 'low'
        self.calc_every_n = 10

    def start(self, seconds=None):
        def delayed_stop():
            time.sleep(seconds)
            print('Timer completed. Disconnecting now...')
            self.stop()

        if len(self.tracking) == 0:
            print('Nothing to track!')
        else:
            self._stream.filter(
                track=self.tracking, languages=self.lang, async=True
                )
#  This does not work due to upstream bug in tweepy 3.5.0. They have fixed it in
#  https://github.com/tweepy/tweepy/pull/783
#            self._stream.filter(
#               track=self.tracking, languages=self.lang, async=True,
#               filter_level=self._filter_level
#               )
        if seconds is not None:
            t = Thread(target=delayed_stop)
            t.start()

    def stop(self):
        self._stream.disconnect()

    def on_data(self, data):
        """
        Note: Due to upstream bug in tweepy for python3, it cannot handle the
        `filter_level` parameter in the `Stream.filter` function. Therefore,
        we'll take care of it here. The problem has been identified and fixed
        by the tweepy team here: https://github.com/tweepy/tweepy/pull/783
        """
        filter_value = {'none': 0, 'low': 1, 'medium': 2}
        value = filter_value[data['filter_level']]

        if value >= filter_value[self._filter_level]:
            self._feels.insert_tweet(data)

    def on_error(self, status):
        pass

    def _intensity(self, tweet):
        t = clean(tweet)
        return SentimentIntensityAnalyzer().polarity_scores(t)['compound']

    @property
    def sentiment(self):
        df = self._feels.queue
        if(len(df)>self.calc_every_n):
            df.sentiment = df.text.apply(self._intensity)
            for row in df.itertuples():
                self._feels.update_tweet(
                    {'id_str': row.id_str, 'sentiment': row.sentiment}
                    )
            df = df.loc[df.sentiment != 0]  # drop rows having 0 sentiment
            g = df.groupby('created_at')
            df['wa'] = df.sentiment*df.followers_count / g.followers_count.transform('sum')
            df = g.wa.sum()
            df = df.sort_index()
            for row in df.iteritems():
                self._sentiment = self._sentiment*0.99 + row[1]*0.01
        return self._sentiment



import sqlite3
import os
import pandas as pd
import logging

class TweetData(object):
    """
    Abstraction of data store.
    """
    def __init__(self, file='feels.sqlite'):
        self._db = file
        if not os.path.isfile(self._db):
            self.make_feels_db(self._db)
        self._debug = False

    @property
    def fields(self):
        conn = sqlite3.connect(self._db)
        c = conn.cursor()
        c.execute("select * from tweets")
        fields=tuple([f[0] for f in c.description])
        c.close()
        return fields

    @property
    def queue(self):
        conn = sqlite3.connect(self._db)
        df = pd.read_sql_query(
            'SELECT * FROM tweets WHERE sentiment is NULL', conn
            )
        return df

    @property
    def all(self):
        conn = sqlite3.connect(self._db)
        df = pd.read_sql_query('SELECT * FROM tweets', conn)
        return df

    def make_feels_db(self, filename='feels.sqlite'):
        conn = sqlite3.connect(filename)
        c = conn.cursor()
        tbl_def = 'CREATE TABLE tweets(\
        id_str          CHARACTER(20)  PRIMARY KEY NOT NULL,\
        text            CHARACTER(140)             NOT NULL,\
        created_at      TEXT                       NOT NULL,\
        coordinates     VARCHAR(20),\
        favorite_count  INTEGER,\
        favorited       VARCHAR(5),\
        lang            VARCHAR(10),\
        place           TEXT,\
        retweet_count   INTEGER,\
        source          TEXT,\
        friends_count   INTEGER,\
        followers_count INTEGER,\
        location        TEXT,\
        sentiment       DOUBLE\
        )'
        c.execute(tbl_def)
        c.close()

    def scrub(self, item):
        if isinstance(item, dict):
            return str(item)
        else:
            return item

    def insert_tweet(self, tweet):
        keys = tuple([k for k in tweet.keys() if k in self.fields])
        vals = tuple([tweet[k] for k in keys])
        ins = '('
        ins = ins + '?,'*(len(vals)-1)
        ins = ins + '?)'
        qry = f'INSERT OR IGNORE INTO tweets {keys} VALUES {ins}'
        try:
            conn = sqlite3.connect(self._db)
            c = conn.cursor()
            c.execute(qry, vals)
            c.close()
            conn.commit()
        except:
            if self._debug:
                logging.warning(f'Failed Query: {qry}, {vals}')

    def update_tweet(self, tweet):
        id_str = tweet['id_str']
        buf = [k for k in tweet.keys() if k!='id_str']
        vals = tuple([tweet[k] for k in buf])
        updt = ''
        while len(buf) > 0:
            cur = buf.pop()
            if len(updt)>0:
                updt = updt + f',{cur}=?'
            else:
                updt = updt + f'{cur}=?'

        qry = f'UPDATE tweets SET {updt} WHERE id_str=?'
        try:
            conn = sqlite3.connect(self._db)
            c = conn.cursor()
            c.execute(qry, vals+(id_str,))
            c.close()
            conn.commit()
        except:
            if self._debug:
                logging.warning(f'Failed Query: {qry}, {vals+(id_str,)}')



import unittest
from tweetfeels import TweetListener
from tweetfeels import Tweet
import json


class Test_Listener(unittest.TestCase):
    def setUp(self):
        self.tweets_data_path = 'test/sample.json'

    def test_listener(self):
        tl = TweetListener(None, None)
        with open(self.tweets_data_path) as tweets_file:
            lines = filter(None, (line.rstrip() for line in tweets_file))
            for line in lines:
                self.assertTrue(tl.on_data(line))

class Test_Tweet(unittest.TestCase):
    def setUp(self):
        self.tweet_file = 'test/tweet.json'
        self.tweet = None
        with open(self.tweet_file) as twt_file:
            lines = list(filter(None, (line.rstrip() for line in twt_file)))
            self.tweet = Tweet(json.loads(lines[0]))

    def test_tweet_keys(self):
        self.assertEqual(self.tweet['followers_count'], 83)
        self.assertEqual(self.tweet['friends_count'], 303)
        self.assertTrue(len(self.tweet)>0)
        self.assertEqual(self.tweet['created_at'], '2017-02-19 19:14:18')

    def test_attributes(self):
        self.assertEqual(len(self.tweet), 29)
        self.assertTrue('followers_count' in self.tweet)



#from distutils.core import setup
from setuptools import setup

setup(name='tweetfeels',
      version='0.1.2',
      description='Real-time sentiment analysis for twitter.',
      author='Thomas Chen',
      author_email='tkchen@gmail.com',
      url='https://github.com/uclatommy/tweetfeels',
      download_url='https://github.com/uclatommy/tweetfeels/tarball/0.1.2',
      packages=['tweetfeels']
     )



import unittest
from tweetfeels import TweetData
from tweetfeels import Tweet
import json
import os


class Test_Data(unittest.TestCase):
    def setUp(self):
        self.tweets_data_path = 'test/sample.json'
        self.db = './test.sqlite'
        self.feels_db = TweetData(self.db)

    def tearDown(self):
        os.remove(self.db)

    def test_file_creation(self):
        self.assertTrue(os.path.exists(self.db))

    def test_fields(self):
        f = self.feels_db.fields
        self.assertTrue(isinstance(f, tuple))
        self.assertTrue(len(f)>=11)

    def test_scrub(self):
        data = {'a':1, 'b':2}
        scrubbed = self.feels_db.scrub(data)
        self.assertTrue(isinstance(scrubbed, str))

    def test_data_operation(self):
        twt = {'created_at': 'Sun Feb 19 19:14:18 +0000 2017',
               'id_str': '833394296418082817',
               'text': 'All the feels!'}
        t = Tweet(twt)
        self.assertEqual(len(t.keys()), 3)
        self.feels_db.insert_tweet(t)
        df = self.feels_db.queue
        self.assertEqual(len(df), 1)
        df.sentiment = 0.9
        for row in df.itertuples():
            self.feels_db.update_tweet(
                {'id_str': row.id_str, 'sentiment': row.sentiment}
                )
        self.assertEqual(len(self.feels_db.queue), 0)
        self.assertEqual(len(self.feels_db.all), 1)
